"""Widget for plugin display
"""

# standard
from pathlib import Path
from typing import Optional

# PyQGIS
from qgis.PyQt import uic
from qgis.PyQt.QtCore import QSize, QUrl
from qgis.PyQt.QtGui import QDesktopServices
from qgis.PyQt.QtWidgets import QPushButton, QWidget

# Project
from oslandia.datamodels.mdl_projects_browser import ProjectObj
from oslandia.toolbelt.plugin_manager import PluginManager
from oslandia.toolbelt.qgis_info import get_qgis_language


class ProjectWidget(QWidget):
    """QWidget to display plugins or product on Oslandia Browser

    :param parent: dialog parent, defaults to None
    :type parent: Optional[QWidget], optional
    """

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        ui_path = Path(__file__).resolve(True).parent / "wdg_cards_projet.ui"
        uic.loadUi(ui_path, self)
        self.tb_about.setOpenExternalLinks(True)
        self.tb_description.setOpenExternalLinks(True)

    def set_display_data(self, project: ProjectObj):

        if project.logo:
            logo = PluginManager().get_project_icon(project.name, project.logo)

            self.set_text_and_font(self.lbl_name, project.name, bold=True, font_size=14)
            self.lbl_logo.setPixmap(logo.pixmap(logo.actualSize(QSize(50, 50))))

        # Description
        if get_qgis_language == "fr" and project.description_fr:
            description = project.description_fr
        else:
            description = project.description_en

        # Set various text fields (Version, About, Tags, Funders)
        self.lbl_version.setText(project.version)
        self.tb_about.setHtml(project.about)
        self.lbl_tags.setText(project.tags)
        self.lbl_funders.setText(project.funders)
        self.tb_description.setHtml(description)

        self.setup_button_with_url(self.btn_doc, project.doc_url)
        self.setup_button_with_url(self.btn_repo, project.code_url)
        self.setup_button_with_url(self.btn_blog, project.blog_link)
        self.setup_button_with_url(self.btn_issue, project.issue_to_be_funded)

    def set_text_and_font(
        self, widget: QWidget, text: str, bold: bool = False, font_size: int = None
    ) -> None:
        """
        Sets the text and font properties of a widget, including optional boldness and font size adjustments.

        :param widget: The widget to set the text and font on (e.g., QLabel, QPushButton)
        :param text: The text to display in the widget
        :param bold: Whether to apply bold formatting to the font (default: False)
        :param font_size: The size of the font to apply (optional)
        :return: None
        """
        widget.setText(text)
        if bold or font_size:
            font = widget.font()
            if bold:
                font.setBold(True)
            if font_size:
                font.setPointSize(font_size)
            widget.setFont(font)

    def setup_button_with_url(self, button: QPushButton, url: str) -> None:
        """
        Configures a QPushButton to open a URL in the default web browser.
        The button is enabled or disabled based on the validity of the URL.

        :param button: Instance of the QPushButton to configure
        :param url: URL to be opened when the button is clicked
        """
        try:
            button.clicked.disconnect()
        except:
            pass
        if url:
            button.setEnabled(True)
        else:
            button.setEnabled(False)
            return

        def open_url():
            QDesktopServices.openUrl(QUrl(url))

        button.clicked.connect(open_url)
